/*
 * JZ4780 HDMI peripheral support.
 *
 * Copyright (C) 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hdmi-jz4780.h>
#include <l4/devices/hw_mmio_register_block.h>

#include <l4/sys/irq.h>
#include <l4/util/util.h>

#include <cstdio>

/*
I2C pins:

HDMI: PF25/SMB4_SDA/DDCSDA, PF24/SMB4_SCK/DDCSCK

See: http://mipscreator.imgtec.com/CI20/hardware/board/ci20_jz4780_v2.0.pdf
*/

enum Regs
{
  // Identification.

  Design_id                 = 0x000,  // DESIGN_ID
  Revision_id               = 0x001,  // REVISION_ID
  Product_id0               = 0x002,  // PRODUCT_ID0
  Product_id1               = 0x003,  // PRODUCT_ID1
  Config_id0                = 0x004,  // CONFIG_ID0
  Config_id1                = 0x005,  // CONFIG_ID1
  Config_id2                = 0x006,  // CONFIG_ID2
  Config_id3                = 0x007,  // CONFIG_ID3

  // Top-level interrupt control.

  Int_mask                  = 0x1ff,  // MUTE

  // Interrupt status and mask for various functions.

  Fc_int_status0            = 0x100,  // FC_STAT0
  Fc_int_status1            = 0x101,  // FC_STAT1
  Fc_int_status2            = 0x102,  // FC_STAT2
  As_int_status             = 0x103,  // AS_STAT0
  Phy_int_status            = 0x104,  // PHY_STAT0
  Cec_int_status            = 0x106,  // CEC_STAT0
  Vp_int_status             = 0x107,  // VP_STAT0
  Ahb_dma_audio_int_status  = 0x109,  // AHBDMAAUD_STAT0

  Fc_int_mask0              = 0x180,  // MUTE_FC_STAT0
  Fc_int_mask1              = 0x181,  // MUTE_FC_STAT1
  Fc_int_mask2              = 0x182,  // MUTE_FC_STAT2
  As_int_mask               = 0x183,  // MUTE_AS_STAT0
  Phy_int_mask              = 0x184,  // MUTE_PHY_STAT0
  Cec_int_mask              = 0x186,  // MUTE_CEC_STAT0
  Vp_int_mask               = 0x187,  // MUTE_VP_STAT0
  Ahb_dma_audio_int_mask    = 0x189,  // MUTE_AHBDMAAUD_STAT0

  // I2C for E-DDC.

  I2c_int_status            = 0x105,  // I2CM_STAT0
  I2c_int_mask              = 0x185,  // MUTE_I2CM_STAT0

  I2c_device_address        = 0x7e00, // I2CM_SLAVE
  I2c_register              = 0x7e01, // I2CM_ADDRESS
  I2c_data_out              = 0x7e02, // I2CM_DATAO
  I2c_data_in               = 0x7e03, // I2CM_DATAI
  I2c_operation             = 0x7e04, // I2CM_OPERATION
  I2c_int_config0           = 0x7e05, // I2CM_INT
  I2c_int_config1           = 0x7e06, // I2CM_CTLINT
  I2c_divider               = 0x7e07, // I2CM_DIV
  I2c_segment_address       = 0x7e08, // I2CM_SEGADDR
  I2c_software_reset        = 0x7e09, // I2CM_SOFTRSTZ
  I2c_segment_pointer       = 0x7e0a, // I2CM_SEGPTR

  // I2C for PHY.

  I2c_phy_int_status        = 0x108,  // I2CMPHY_STAT0
  I2c_phy_int_mask          = 0x188,  // MUTE_I2CMPHY_STAT0

  I2c_phy_int_config0       = 0x3027, // PHY_I2CM_INT_ADDR
  I2c_phy_int_config1       = 0x3028, // PHY_I2CM_CTLINT_ADDR
};

// Identification values.

enum Product_id_values : unsigned
{
  Product_id0_transmitter   = 0xa0, // PRODUCT_ID0_HDMI_TX 

  Product_id1_hdcp          = 0xc0, // PRODUCT_ID1_HDCP 
  Product_id1_receiver      = 0x02, // PRODUCT_ID1_HDMI_RX 
  Product_id1_transmitter   = 0x01, // PRODUCT_ID1_HDMI_TX
};

// Configuration values.

enum Config_id_values : unsigned
{
  Config_id0_i2s                  = 0x10, // CONFIG0_I2S
  Config_id0_cec                  = 0x02, // CONFIG0_CEC

  Config_id1_ahb                  = 0x01, // CONFIG1_AHB

  Config2_dwc_hdmi_tx_phy         = 0x00, // DWC_HDMI_TX_PHY
  Config2_dwc_mhl_phy_heac        = 0xb2, // DWC_MHL_PHY_HEAC
  Config2_dwc_mhl_phy             = 0xc2, // DWC_MHL_PHY
  Config2_dwc_hdmi_3d_tx_phy_heac = 0xe2, // DWC_HDMI_3D_TX_PHY_HEAC
  Config2_dwc_hdmi_3d_tx_phy      = 0xf2, // DWC_HDMI_3D_TX_PHY
  Config2_dwc_hdmi20_tx_phy       = 0xf3, // DWC_HDMI20_TX_PHY
  Config2_vendor_phy              = 0xfe, // VENDOR_PHY

  Config_id3_ahb_audio_dma        = 0x02, // CONFIG3_AHBAUDDMA
  Config_id3_gp_audio             = 0x01, // CONFIG3_GPAUD
};

// Status and mask bits.

enum Int_mask_bits : unsigned
{
  Int_mask_wakeup = 0x2,
  Int_mask_all    = 0x1,
};

enum I2c_int_status_bits : unsigned
{
  I2c_int_status_done     = 0x2,
  I2c_int_status_error    = 0x1,
};

// I2C operation bits.

enum I2c_operation_bits : unsigned
{
  I2c_operation_write         = 0x10,
  I2c_operation_segment_read  = 0x2,
  I2c_operation_read          = 0x1,
};

// Interrupt configuration bits.

enum I2c_int_config0_bits : unsigned
{
  I2c_int_config_done_polarity  = 0x8,
  I2c_int_config_done_mask      = 0x4,
};

enum I2c_int_config1_bits : unsigned
{
  I2c_int_config_nack_polarity  = 0x80,
  I2c_int_config_nack_mask      = 0x40,
  I2c_int_config_arb_polarity   = 0x8,
  I2c_int_config_arb_mask       = 0x4,
};



// Initialise the HDMI peripheral.

Hdmi_jz4780_chip::Hdmi_jz4780_chip(l4_addr_t start, l4_addr_t end,
                                   l4_cap_idx_t irq)
: _start(start), _end(end), _irq(irq)
{
  // 8-bit registers with 2-bit address shifting.

  _regs = new Hw::Mmio_register_block<8>(start, 2);

  _segment_read = false;
  _device_register = 0;

  get_identification();
  int_init();
  i2c_init();
}

void Hdmi_jz4780_chip::get_identification()
{
  _version = (_regs[Design_id] << 8) | _regs[Revision_id];
}

void Hdmi_jz4780_chip::get_version(uint8_t *major, uint16_t *minor)
{
  *major = _version >> 12;
  *minor = _version & 0xfff;
}

void Hdmi_jz4780_chip::int_init()
{
  // Disable interrupts.

  _regs[Int_mask] = _regs[Int_mask] | (Int_mask_wakeup | Int_mask_all);

  // Mask all interrupts.

  _regs[Fc_int_mask0] = 0xff;
  _regs[Fc_int_mask1] = 0xff;
  _regs[Fc_int_mask2] = 0xff;
  _regs[As_int_mask] = 0xff;
  _regs[Phy_int_mask] = 0xff;
  _regs[I2c_int_mask] = 0xff;
  _regs[Cec_int_mask] = 0xff;
  _regs[Vp_int_mask] = 0xff;
  _regs[I2c_phy_int_mask] = 0xff;
  _regs[Ahb_dma_audio_int_mask] = 0xff;

  // Enable interrupts.

  _regs[Int_mask] = _regs[Int_mask] & ~(Int_mask_wakeup | Int_mask_all);
}

void Hdmi_jz4780_chip::i2c_init()
{
  // Set PHY interrupt priorities.

  _regs[I2c_phy_int_config0] = I2c_int_config_done_polarity;
  _regs[I2c_phy_int_config1] = I2c_int_config_nack_polarity |
                               I2c_int_config_arb_polarity;

  // Software reset.

  _regs[I2c_software_reset] = 0;

  // Standard mode (100kHz).

  _regs[I2c_divider] = 0;

  // Set interrupt polarities.

  _regs[I2c_int_config0] = I2c_int_config_done_polarity;
  _regs[I2c_int_config1] = I2c_int_config_nack_polarity |
                           I2c_int_config_arb_polarity;

  // Clear and mask/mute interrupts.

  _regs[I2c_int_status] = I2c_int_status_done | I2c_int_status_error;
  _regs[I2c_int_mask] = I2c_int_status_done | I2c_int_status_error;
}

long Hdmi_jz4780_chip::i2c_wait()
{
  long err;

  // Wait for around 1s.

  l4_msgtag_t tag = l4_irq_receive(_irq, L4_IPC_NEVER);

  err = l4_ipc_error(tag, l4_utcb());
  if (err)
    return err;

  // Test for an error condition.

  if (_regs[I2c_int_status] & I2c_int_status_error)
    return -L4_EIO;

  _regs[I2c_int_status] = _regs[I2c_int_status] | I2c_int_status_done;

  return L4_EOK;
}

int Hdmi_jz4780_chip::i2c_read(uint8_t *buf, unsigned int length)
{
  unsigned int i;
  long err;

  // Clear interrupts.

  _regs[I2c_int_mask] = 0;

  for (i = 0; i < length; i++)
  {
    // Increment the device register.

    _regs[I2c_register] = _device_register++;
    _regs[I2c_operation] = _segment_read ? I2c_operation_segment_read
                                         : I2c_operation_read;

    // Wait and then read.

    err = i2c_wait();
    if (err)
      break;

    buf[i] = _regs[I2c_data_in];
  }

  // Mask interrupts again.

  _regs[I2c_int_mask] = I2c_int_status_done | I2c_int_status_error;

  return i;
}

void Hdmi_jz4780_chip::i2c_set_address(uint8_t address)
{
  _regs[I2c_device_address] = address;
  _segment_read = false;
  i2c_set_register(0);
}

void Hdmi_jz4780_chip::i2c_set_segment(uint8_t segment)
{
  _regs[I2c_segment_address] = 0x30;
  _regs[I2c_segment_pointer] = segment;
  _segment_read = true;
  i2c_set_register(0);
}

void Hdmi_jz4780_chip::i2c_set_register(uint8_t device_register)
{
  _device_register = device_register;
}



// C language interface functions.

void *jz4780_hdmi_init(l4_addr_t start, l4_addr_t end, l4_cap_idx_t irq)
{
  return (void *) new Hdmi_jz4780_chip(start, end, irq);
}

void jz4780_hdmi_get_version(void *hdmi, uint8_t *major, uint16_t *minor)
{
  static_cast<Hdmi_jz4780_chip *>(hdmi)->get_version(major, minor);
}

int jz4780_hdmi_i2c_read(void *hdmi, uint8_t *buf, unsigned int length)
{
  return static_cast<Hdmi_jz4780_chip *>(hdmi)->i2c_read(buf, length);
}

void jz4780_hdmi_i2c_set_address(void *hdmi, uint8_t address)
{
  static_cast<Hdmi_jz4780_chip *>(hdmi)->i2c_set_address(address);
}

void jz4780_hdmi_i2c_set_segment(void *hdmi, uint8_t segment)
{
  static_cast<Hdmi_jz4780_chip *>(hdmi)->i2c_set_segment(segment);
}

void jz4780_hdmi_i2c_set_register(void *hdmi, uint8_t device_register)
{
  static_cast<Hdmi_jz4780_chip *>(hdmi)->i2c_set_register(device_register);
}
